@@ -27,11 +27,12 @@ from .temporary_patches.common import UNSLOTH_ENABLE_LOGGING, logger
 
 
 try:
-    from transformers.integrations.mxfp4 import convert_moe_packed_tensors
+    from transformers.integrations.mxfp4 import convert_moe_packed_tensors, convert_moe_packed_tensors_cpu
 except (ImportError, ModuleNotFoundError):
     # Provide a fallback or a clear error if the function isn't available
     # when not using mxfp4.
-    convert_moe_packed_tensors = None
+    convert_moe_packed_tensors     = None
+    convert_moe_packed_tensors_cpu = None
 
 
 MODEL_CARD = \
@@ -361,9 +362,44 @@ pass
 import torch
 import gc
 import time
+import safetensors
+import json
+import mmap
+# Mapping from BF16 to torch.blfloat16 etc
+try:
+    SAFETENSORS_DTYPES = safetensors.torch._TYPES
+except:
+    logger.info("Unsloth: `safetensors.torch._TYPES` does not exist. Will set to our default version")
+    SAFETENSORS_DTYPES = {
+        'F64': torch.float64,
+        'F32': torch.float32,
+        'F16': torch.float16,
+        'BF16': torch.bfloat16,
+        'I64': torch.int64,
+        'I32': torch.int32,
+        'I16': torch.int16,
+        'I8': torch.int8,
+        'U8': torch.uint8,
+        'BOOL': torch.bool,
+        'F8_E4M3': torch.float8_e4m3fn,
+        'F8_E5M2': torch.float8_e5m2,
+        'U64': torch.uint64,
+        'U32': torch.uint32,
+        'U16': torch.uint16,
+    }
+pass
 
 @torch.inference_mode
-def _merge_and_overwrite_lora(save_directory, filename, lora_weights, output_dtype, model_class_name, base_model_is_quantized=False, quant_type=None):
+def _merge_and_overwrite_lora(
+    save_directory,
+    filename,
+    lora_weights,
+    output_dtype,
+    model_class_name,
+    base_model_is_quantized = False,
+    quant_type = None,
+    overwrite = False, # Overwrites current safetensor file
+):
     # All Unsloth Zoo code licensed under LGPLv3
     # Merges LoRA and overwrites the safetensors file it was merged to
     filename_original = os.path.join(save_directory, filename)  # Original file path
@@ -375,22 +411,68 @@ def _merge_and_overwrite_lora(save_directory, filename, lora_weights, output_dty
 
     # Convert lora_weights to safetensor format
     converted_lora_weights = _convert_lora_keys_to_safetensor_format(
-        lora_weights, [], model_class_name=model_class_name)
+        lora_weights,
+        [],
+        model_class_name = model_class_name,
+    )
 
-    with safe_open(filename_original, framework="pt", device="cpu") as file:  # Open original file for reading
+    # Open original file for reading
+    raw_pointer = None
+    mm = None
+    raw_safetensor_keys = None
+    header_metadata = None
+    length_of_header = 0
+    # Only if overwriting
+    if overwrite:
+        raw_pointer = open(filename_original, "r+b")
+        # Memory map entire file
+        mm = mmap.mmap(raw_pointer.fileno(), length = 0, access = mmap.ACCESS_WRITE)
+        if mm is None:
+            raw_pointer.flush()
+            raw_pointer.close()
+            overwrite = False
+        else:
+            # Get metadata which is a dict
+            # https://gist.github.com/Narsil/3edeec2669a5e94e4707aa0f901d2282#file-pure_torch-py-L8
+            length_of_header = int.from_bytes(mm.read(8), "little")
+            header_metadata = json.loads(mm.read(length_of_header))
+            mm.seek(0)
+            raw_safetensor_keys = [k for k in header_metadata.keys() if k != "__metadata__"]
+    pass
+
+    with safe_open(filename_original, framework = "pt", device = "cpu") as file:
         safetensor_keys = list(file.keys())
 
+        if overwrite and raw_safetensor_keys is not None:
+            # Get keys and compare
+            difference = set(safetensor_keys) ^ set(raw_safetensor_keys)
+            if len(difference) != 0:
+                logger.info("Unsloth: Disabling disk space efficient overwriting and switching to less efficient temporary file approach.")
+                # Disable overwrite
+                overwrite = False
+                mm.flush(); os.fsync(raw_pointer.fileno()); mm.close(); mm = None
+                raw_pointer.flush(); raw_pointer.close(); raw_pointer = None
+        pass
+
         # Update converted_lora_weights with actual safetensor keys
         converted_lora_weights = _convert_lora_keys_to_safetensor_format(
-            lora_weights, safetensor_keys, model_class_name=model_class_name)
+            lora_weights,
+            safetensor_keys,
+            model_class_name = model_class_name,
+        )
 
         # Set to track mxfp4 keys that have already been processed
         processed_mxfp4_keys = set()
 
-        for key in safetensor_keys:
+        for iteration, key in enumerate(safetensor_keys):
             if key in processed_mxfp4_keys:
                 continue
 
+            # FORCE memory cleanup before processing each tensor
+            if not overwrite:
+                torch.cuda.empty_cache()
+                torch.cuda.synchronize()
+
             W = None
             output_key = key
             action_logged = False
@@ -403,14 +485,47 @@ def _merge_and_overwrite_lora(save_directory, filename, lora_weights, output_dty
                     if key.endswith("_blocks"):
                         if convert_moe_packed_tensors is None:
                             raise ImportError("MXFP4 dequantization is required, but `convert_moe_packed_tensors` could not be imported.")
+
                         base_name = key[:-len("_blocks")]
                         scales_key = base_name + "_scales"
                         output_key = base_name # Correct naming without .weight
                         if scales_key not in safetensor_keys:
                             warnings.warn(f"Found mxfp4 tensor {key} but missing its scales tensor {scales_key}. Skipping.")
                             continue
+
                         blocks_tensor, scales_tensor = file.get_tensor(key), file.get_tensor(scales_key)
-                        W = convert_moe_packed_tensors(blocks_tensor, scales_tensor).transpose(1, 2).contiguous()
+
+                        torch.cuda.synchronize() # Wait for previous operations to complete
+                        torch.cuda.empty_cache()
+
+                        # Determine optimal device and chunk size for mxfp4 dequantization
+                        device_type, device_id, rows_per_chunk = _choose_mxfp4_processing_strategy(
+                            blocks_tensor, scales_tensor
+                        )
+
+                        # Apply dequantization with optimal parameters
+                        if device_type == 'cpu':
+                            # Use CPU-optimized version
+                            try:
+                                from transformers.integrations.mxfp4 import convert_moe_packed_tensors_cpu
+                                W = convert_moe_packed_tensors_cpu(
+                                    blocks_tensor, scales_tensor, rows_per_chunk=rows_per_chunk
+                                ).transpose(1, 2).contiguous()
+                                if UNSLOTH_ENABLE_LOGGING:
+                                    logger.info(f"[DEBUG] Using CPU dequantization for {base_name} with {rows_per_chunk:,} rows per chunk")
+                            except ImportError:
+                                # Fallback to original function
+                                W = convert_moe_packed_tensors(
+                                    blocks_tensor, scales_tensor, rows_per_chunk=rows_per_chunk
+                                ).transpose(1, 2).contiguous()
+                        else:
+                            # Use GPU version (original or patched)
+                            W = convert_moe_packed_tensors(
+                                blocks_tensor, scales_tensor, rows_per_chunk=rows_per_chunk
+                            ).transpose(1, 2).contiguous()
+                            if UNSLOTH_ENABLE_LOGGING:
+                                logger.info(f"[DEBUG] Using GPU dequantization for {base_name} with {rows_per_chunk:,} rows per chunk")
+
                         processed_mxfp4_keys.add(key); processed_mxfp4_keys.add(scales_key)
 
                         lora_stats = converted_lora_weights.get(base_name, None)
@@ -440,43 +555,85 @@ def _merge_and_overwrite_lora(save_directory, filename, lora_weights, output_dty
             if W is not None and lora_stats is not None and hasattr(lora_stats, 'lora_A') and lora_stats.lora_A is not None:
                 if not action_logged:
                     count += 1
-                    W = _merge_lora(W, lora_stats, output_key)  # Assume _merge_lora is defined elsewhere
-                    action_logged=True
+                    W = _merge_lora(W, lora_stats, output_key) # Assume _merge_lora is defined elsewhere
+                    action_logged = True
 
             if W is None:
                 continue
 
-            with tempfile.NamedTemporaryFile(suffix=".pt", delete=False) as temp_file:
-                temp_filename = temp_file.name
-                # Save the merged tensor to a unique temp file
-                torch.save(W.to(output_dtype), temp_filename, pickle_module=pickle, pickle_protocol=pickle.HIGHEST_PROTOCOL)
-                # Load it back as a memory-mapped object. The OS will manage paging this from disk.
-                W = torch.load(temp_filename, map_location="cpu", mmap=True, weights_only=False)
-
-                # Clean up the temporary pickle file immediately after mmaping
-            try:
-                os.remove(temp_filename)
-            except OSError:
-                # On Windows, the mmap might keep a handle. The OS will clean it up.
+            # Overwriting logic
+            success = False
+            if overwrite and mm is not None and header_metadata is not None:
+                key_metadata = header_metadata[output_key]
+                index_L, index_R = key_metadata["data_offsets"]
+                # dtype = SAFETENSORS_DTYPES[key_metadata["dtype"]]
+                # shape = key_metadata["shape"]
+                index_L += 8 + length_of_header # Indexing via [] uses seek(0)
+                index_R += 8 + length_of_header # Indexing via [] uses seek(0)
+                # array = mm[index_L : index_R]
+                # Transform tensor to numpy bytes format
+                numpy_view = W.to(output_dtype).view(torch.uint8).contiguous().ravel()
+                numpy_view = numpy_view.to("cpu", non_blocking = True).numpy()
+                try:
+                    # mm[index_L : index_R] = memoryview(numpy_view)
+                    mm.seek(index_L)
+                    mm.write(numpy_view.tobytes())
+                    success = True
+                    del W
+                    del numpy_view
+                except Exception as e:
+                    logger.info(f"Failed overwriting safetensor file at `{output_key}` with error = {str(e)}. Switching to slower version.")
+                    if numpy_view.nbytes != (index_R - index_L):
+                        logger.info(f"Fast saving got incorrect elements with `numpy_view={numpy_view.nbytes}, mmap={index_R-index_L}`")
+                    # Disable overwrite
+                    overwrite = False
+                    mm.flush(); os.fsync(raw_pointer.fileno()); mm.close(); mm = None
+                    raw_pointer.flush(); raw_pointer.close(); raw_pointer = None
+            pass
+            if not success:
+                # Slower and uses more disk space
+                with tempfile.NamedTemporaryFile(suffix = ".pt", delete = False) as temp_file:
+                    temp_filename = temp_file.name
+                    # Save the merged tensor to a unique temp file
+                    torch.save(W.to(output_dtype), temp_filename, pickle_module = pickle, pickle_protocol = pickle.HIGHEST_PROTOCOL)
+                    # Load it back as a memory-mapped object. The OS will manage paging this from disk.
+                    W = torch.load(temp_filename, map_location = "cpu", mmap = True, weights_only = False)
                 pass
-
-
-            tensors[output_key] = W
-
-            # Free up VRAM after each merge
-            torch.cuda.empty_cache()
+                # Clean up the temporary pickle file immediately after mmaping
+                try:
+                    os.remove(temp_filename)
+                except OSError:
+                    # On Windows, the mmap might keep a handle. The OS will clean it up.
+                    pass
+                # Save to tensors output for saving in the future
+                tensors[output_key] = W
+                # Free up VRAM after each merge
+                torch.cuda.empty_cache()
+            pass
+        pass
+    pass
 
     # CRITICAL: Force cleanup to release file handles on Windows
-    if os.name == 'nt':
+    if os.name == "nt":
         gc.collect()
         time.sleep(0.1)  # Give Windows a moment to release file handles
 
+    # Final cleanup for overwrite
+    if overwrite:
+        overwrite = False
+        mm.flush(); os.fsync(raw_pointer.fileno()); mm.close(); mm = None
+        raw_pointer.flush(); raw_pointer.close(); raw_pointer = None
+        gc.collect()
+        # FAST return if overwriting
+        return count
+    pass
+
     # Create a temporary file in the same directory for atomic rename
-    with tempfile.NamedTemporaryFile(suffix=".safetensors", dir=save_directory, delete=False) as tmpfile:
+    with tempfile.NamedTemporaryFile(suffix = ".safetensors", dir = save_directory, delete = False) as tmpfile:
         temp_filename_safetensors = tmpfile.name
 
-    save_file(tensors, temp_filename_safetensors, metadata={"format": "pt"})  # Save to the temporary safetensors file
-
+    save_file(tensors, temp_filename_safetensors, metadata = {"format": "pt"})  # Save to the temporary safetensors file
+    del tensors
     # Replace the temporary file with the original file
     try:
         os.replace(temp_filename_safetensors, filename_original)  # Attempt atomic rename
@@ -505,9 +662,10 @@ def _merge_and_overwrite_lora(save_directory, filename, lora_weights, output_dty
             os.remove(temp_filename_safetensors)
         except:
             pass
-
+    pass
     return count
 pass
+
 from huggingface_hub import (
     split_state_dict_into_shards_factory,
     get_torch_storage_size,
@@ -698,6 +856,7 @@ def merge_and_overwrite_lora(
         model_name = get_model_name(model.config._name_or_path, load_in_4bit = False)
     except:
         model_name = model.config._name_or_path
+        model_name = model_name.removesuffix("-unsloth-bnb-4bit").removesuffix("-bnb-4bit")
 
     final_model_name, is_local_path, source_info, base_model_is_quantized, quant_type = determine_base_model_source(model_name, token)
     if base_model_is_quantized and (quant_type == "nf4" or quant_type == "fp4") and save_method== "merged_16bit":
@@ -862,6 +1021,15 @@ def merge_and_overwrite_lora(
             save_directory = save_directory,
             state_dict = {},
         )
+        # Remove any weight files that shouldn't have been saved (transformers 4.56.0 bug)
+        import glob
+        weight_files = glob.glob(os.path.join(save_directory, "*.bin")) + \
+                       glob.glob(os.path.join(save_directory, "*.safetensors"))
+
+        for weight_file in weight_files:
+            os.remove(weight_file)
+            print(f"DEBUG: Removed incorrectly saved weight file: {os.path.basename(weight_file)}")
+
         _remove_quantization_config(config_path = Path(save_directory) / "config.json")
         # Remove the quantization_config in the config.json file if it exists,
     # as we are exporting the model in 16-bit format.
@@ -870,7 +1038,6 @@ def merge_and_overwrite_lora(
     if push_to_hub:
         upload_items()
 
-
     # Step 3: Conditional index handling
     _hf_cache_dir = _get_hf_cache_dir()
     copied_all_from_cache = False
@@ -888,7 +1055,12 @@ def merge_and_overwrite_lora(
         else:
             # Download from HF
             if "model.safetensors.index.json" in [f for f in safe_tensor_index_files]:
-                snapshot_download(repo_id=model_name, local_dir=save_directory, allow_patterns=["model.safetensors.index.json"])
+                snapshot_download(
+                    repo_id = model_name,
+                    local_dir = save_directory,
+                    allow_patterns = ["model.safetensors.index.json"],
+                    local_dir_use_symlinks = False,
+                )
 
         if push_to_hub and safe_tensor_index_files:
             upload_items("model.safetensors.index.json")
@@ -910,6 +1082,7 @@ def merge_and_overwrite_lora(
             repo_id = model_name,
             local_dir = save_directory,
             allow_patterns = safe_tensor_index_files + safetensors_list,
+            local_dir_use_symlinks = False,
         )
 
 
@@ -940,7 +1113,8 @@ def merge_and_overwrite_lora(
             output_dtype = output_dtype,
             model_class_name = find_lora_base_model(model).__class__.__name__,
             base_model_is_quantized = base_model_is_quantized,
-            quant_type=quant_type,
+            quant_type = quant_type,
+            overwrite = False, # Faster overwriting logic
         )
         torch.cuda.empty_cache()
         if low_disk_space_usage and push_to_hub:
@@ -1029,7 +1203,7 @@ def _try_copy_all_from_cache(
     all_found = True
     for filename in filenames_to_check:
         try:
-            cached_path_str = hf_hub_download(repo_id=repo_id, filename=filename, local_files_only=True)
+            cached_path_str = hf_hub_download(repo_id = repo_id, filename = filename, local_files_only = True)
             cached_paths_map[filename] = Path(cached_path_str) # Store Path for checking
         except LocalEntryNotFoundError:
             print(f"Cache check failed: {filename} not found in local cache.") # Verbose
@@ -1555,12 +1729,11 @@ def check_model_quantization_status(model_name_or_path, token=None):
     # HF repo
     else:
         try:
-            from huggingface_hub import hf_hub_download
             config_path = hf_hub_download(
-                repo_id=model_name_or_path,
-                filename="config.json",
-                cache_dir=None,
-                token=token
+                repo_id = model_name_or_path,
+                filename = "config.json",
+                cache_dir = None,
+                token = token
             )
             with open(config_path, 'r', encoding="utf-8") as f:
                 config = json.load(f)
@@ -1629,6 +1802,218 @@ def determine_base_model_source(model_name, token=None):
             raise ValueError(f"Model {model_name} not found locally or on HuggingFace")
 pass
 
+def get_memory_stats():
+    """Get current memory statistics for CPU and GPU"""
+    stats = {}
+    import psutil
+
+    # CPU Memory
+    cpu_mem = psutil.virtual_memory()
+    stats['cpu'] = {
+        'total': cpu_mem.total,
+        'available': cpu_mem.available,
+        'used': cpu_mem.used,
+        'percent': cpu_mem.percent,
+        'free': cpu_mem.available,  # Available is more accurate than free
+    }
+
+    # GPU Memory (for each GPU)
+    stats['gpus'] = []
+    if torch.cuda.is_available():
+        for i in range(torch.cuda.device_count()):
+            gpu_mem = torch.cuda.mem_get_info(i)
+            total = gpu_mem[1]
+            free = gpu_mem[0]
+            stats['gpus'].append({
+                'device_id': i,
+                'name': torch.cuda.get_device_name(i),
+                'total': total,
+                'free': free,
+                'used': total - free,
+                'percent': ((total - free) / total) * 100 if total > 0 else 0
+            })
+
+    return stats
+pass
+
+def format_bytes(bytes_value):
+    """Convert bytes to human readable format"""
+    for unit in ['B', 'KB', 'MB', 'GB', 'TB']:
+        if bytes_value < 1024.0:
+            return f"{bytes_value:.2f} {unit}"
+        bytes_value /= 1024.0
+    return f"{bytes_value:.2f} PB"
+pass
+
+def calculate_combined_score(speed_score, chunk_size):
+    # Normalize chunk size to 0-1 scale (assuming max reasonable chunk is 100M)
+    chunk_factor = min(1.0, chunk_size / (100 * 1024 * 1024))
+    # Weight: 60% device speed, 40% chunk efficiency
+    return speed_score * 0.6 + chunk_factor * 10 * 0.4  # Scale chunk factor to match speed range
+pass
+
+def _choose_mxfp4_processing_strategy(blocks_tensor, scales_tensor):
+    """
+    Choose optimal device and chunk size for mxfp4 dequantization based on available memory.
+    """
+    import math
+
+    # Calculate tensor dimensions
+    *prefix_shape, G, B = blocks_tensor.shape
+    rows_total = math.prod(prefix_shape) * G
+
+    # Estimate memory requirements
+    #base_memory_per_row = B * 21 if B else 128 * 21
+    base_memory_per_row = B * 35 if B else 128 * 35
+    input_size = blocks_tensor.numel() + scales_tensor.numel() * 4
+    output_size = rows_total * B * 2 * 2
+    persistent_memory = input_size + output_size
+
+    # Device-specific safety factors
+    GPU_SAFETY_FACTOR = 0.75  # GPUs can handle higher utilization
+    CPU_SAFETY_FACTOR = 0.75  # CPUs need more headroom for OS and other processes
+
+    def calculate_safe_usable_memory(free_memory, safety_factor):
+        # Option 1: What we can use from reported free memory (accounting for fragmentation)
+        usable_from_free = free_memory * safety_factor
+
+        return usable_from_free
+
+
+    def calculate_optimal_chunk_size(safe_usable_memory):
+        """Calculate the largest chunk size that fits in the safe usable memory"""
+        temp_memory_budget = safe_usable_memory - persistent_memory
+        if temp_memory_budget <= 0:
+            return None
+
+        max_chunk_from_memory = int(temp_memory_budget // base_memory_per_row)
+        optimal_chunk = min(rows_total, max_chunk_from_memory)
+
+        if optimal_chunk < 1024:
+            return None
+
+        return optimal_chunk
+
+    stats = get_memory_stats()
+    suitable_strategies = []
+
+    # Check GPU strategies first (preferred for speed)
+    for gpu in stats['gpus']:
+        safe_usable_memory = calculate_safe_usable_memory(
+            free_memory=gpu['free'],
+            safety_factor=GPU_SAFETY_FACTOR,
+        )
+        chunk_size = calculate_optimal_chunk_size(safe_usable_memory)
+
+        if chunk_size:
+            temp_memory = min(chunk_size, rows_total) * base_memory_per_row
+            total_memory_needed = persistent_memory + temp_memory
+
+            combined_score = calculate_combined_score(3.0, chunk_size)
+
+            suitable_strategies.append({
+                'device_type': 'cuda',
+                'device_id': gpu['device_id'],
+                'rows_per_chunk': chunk_size,
+                'available_memory': gpu['free'] * GPU_SAFETY_FACTOR,
+                'total_memory': gpu['total'],
+                'safe_usable_memory': safe_usable_memory,
+                'needed_memory': total_memory_needed,
+                'speed_score': 3.0,
+                'efficiency_score': chunk_size,
+                'safety_factor': GPU_SAFETY_FACTOR,
+                'memory_utilization': total_memory_needed / safe_usable_memory,
+                'combined_score': combined_score,
+            })
+
+    # Check CPU strategy
+    cpu_safe_usable_memory = calculate_safe_usable_memory(
+        free_memory=stats['cpu']['available'],
+        safety_factor=CPU_SAFETY_FACTOR,
+    )
+    cpu_chunk_size = calculate_optimal_chunk_size(cpu_safe_usable_memory)
+
+    if cpu_chunk_size:
+        temp_memory = min(cpu_chunk_size, rows_total) * base_memory_per_row
+        total_memory_needed = persistent_memory + temp_memory
+        combined_score = calculate_combined_score(1.0, cpu_chunk_size)  # For CPU
+        suitable_strategies.append({
+            'device_type': 'cpu',
+            'device_id': None,
+            'rows_per_chunk': cpu_chunk_size,
+            'available_memory': stats['cpu']['available'] * CPU_SAFETY_FACTOR,
+            'total_memory': stats['cpu']['total'],
+            'safe_usable_memory': cpu_safe_usable_memory,
+            'needed_memory': total_memory_needed,
+            'speed_score': 1.0,
+            'efficiency_score': cpu_chunk_size,
+            'safety_factor': CPU_SAFETY_FACTOR,
+            'fragmentation_factor': 1.0,
+            'memory_utilization': total_memory_needed / cpu_safe_usable_memory,
+            'combined_score': combined_score,
+        })
+
+    if suitable_strategies:
+
+        # Sort by combined score
+        suitable_strategies.sort(key=lambda x: x['combined_score'], reverse=True)
+
+        best = suitable_strategies[0]
+
+        if UNSLOTH_ENABLE_LOGGING:
+            logger.info(
+                f"[MXFP4] Selected {best['device_type']}:{best['device_id'] or ''} "
+                f"with {best['rows_per_chunk']:,} rows per chunk "
+                f"(safety factor: {best['safety_factor']:.0%}, "
+                f"safe memory utilization: {best['memory_utilization']:.1%}) "
+                f"- Need: {format_bytes(best['needed_memory'])}, "
+                f"Available: {format_bytes(best['available_memory'])}"
+            )
+
+        return (best['device_type'], best['device_id'], best['rows_per_chunk'])
+
+    # Fallback: find device with most memory and use minimal chunk
+    fallback_options = []
+
+    # Add CPU fallback
+    fallback_options.append({
+        'device_type': 'cpu',
+        'device_id': None,
+        'available': stats['cpu']['available'] * CPU_SAFETY_FACTOR,
+        'total_available': stats['cpu']['available']
+    })
+
+    # Add GPU fallbacks
+    for gpu in stats['gpus']:
+        fallback_options.append({
+            'device_type': 'cuda',
+            'device_id': gpu['device_id'],
+            'available': gpu['free'] * GPU_SAFETY_FACTOR,
+            'total_available': gpu['free']
+        })
+
+    # Sort by available memory (after safety factor)
+    fallback_options.sort(key=lambda x: x['available'], reverse=True)
+    best_fallback = fallback_options[0]
+
+    # Calculate minimal safe chunk size for fallback
+    remaining_memory = best_fallback['available'] - persistent_memory
+    if remaining_memory > 0:
+        fallback_chunk_size = max(1024, min(8192, int(remaining_memory // base_memory_per_row), rows_total))
+    else:
+        fallback_chunk_size = min(1024, rows_total)
+
+    warnings.warn(
+        f"[MXFP4] Insufficient memory for optimal processing on any device. "
+        f"Using {best_fallback['device_type']}:{best_fallback['device_id'] or ''} "
+        f"with minimal chunks ({fallback_chunk_size:,}). "
+        f"Available: {format_bytes(best_fallback['total_available'])}, "
+        f"Required: {format_bytes(persistent_memory)}. "
+        f"Processing will be slow."
+    )
+
+    return (best_fallback['device_type'], best_fallback['device_id'], fallback_chunk_size)
+pass
 # Unsloth Zoo - Utilities for Unsloth
 # Copyright 2023-present Daniel Han-Chen, Michael Han-Chen & the Unsloth team. All rights reserved.
 #
