@@ -20,10 +20,14 @@ import torch
 import torch.nn as nn
 import torch.nn.functional as F
 import inspect
-import textwrap
 from .common import TEMPORARY_PATCHES, torch_compile
+from importlib.metadata import version as importlib_version
+from ..utils import Version
+transformers_version = Version(importlib_version("transformers"))
 from .utils import (
     patch_function,
+    patch_function_past_key_values,
+    dedent,
     KWARGS_TYPE,
     raise_error,
     logger,
@@ -107,7 +111,7 @@ def patch_gpt_oss():
     except Exception as e:
         return raise_error("transformers.integrations.mxfp4", e)
 
-    def swizzle_mxfp4(w, w_scale):
+    def swizzle_mxfp4(w, w_scale, *args, **kwargs):
         from triton_kernels import tensor, tensor_details
         FP4, convert_layout, wrap_torch_tensor = (
             tensor.FP4,
@@ -133,7 +137,7 @@ def patch_gpt_oss():
         # w_scale = convert_layout(wrap_torch_tensor(w_scale), scale_layout, **scale_layout_opts)
         w_scale = convert_layout(wrap_torch_tensor(w_scale), StridedLayout)
         return w, w_scale
-    patch_function(transformers.integrations.mxfp4, "swizzle_mxfp4", swizzle_mxfp4)
+    patch_function(transformers.integrations.mxfp4, "swizzle_mxfp4", swizzle_mxfp4, match_level = "relaxed")
 
     class Mxfp4GptOssExperts_Training(torch.autograd.Function):
         @staticmethod
@@ -208,7 +212,7 @@ def patch_gpt_oss():
             del Wd_T
             # 3) activation derivative
             g1 = swiglu_torch_backward(pre_act, alpha, limit, g1)
-            # 4) g1 · Wuᵀ  
+            # 4) g1 · Wuᵀ
             Wu_T = ctx.self_class.gate_up_proj.data.swapaxes(1, 2).transpose(1, 2).contiguous().transpose(1, 2) # (E, 2*d_ff, d_model)
             dx_exp = matmul_ogs(g1, Wu_T, None, ctx.routing_data, scatter_indx=ctx.gather_idx)
             del Wu_T
@@ -325,7 +329,7 @@ def patch_gpt_oss():
     except Exception as e:
         return raise_error("transformers.integrations.tensor_parallel.shard_and_distribute_module", e)
 
-    def load_and_swizzle_mxfp4(module, param_name, param_value, target_device, **kwargs):
+    def load_and_swizzle_mxfp4(module, param_name, param_value, target_device, *args, **kwargs):
         model = kwargs.get("model", None)
         empty_param = kwargs.get("empty_param", None)
         casting_dtype = kwargs.get("casting_dtype", None)
@@ -388,7 +392,7 @@ def patch_gpt_oss():
                     # setattr(module, blocks_attr, torch.nn.Parameter(triton_weight_tensor.storage.data, requires_grad=False))
                     del blocks
     pass
-    patch_function(transformers.integrations.mxfp4, "load_and_swizzle_mxfp4", load_and_swizzle_mxfp4)
+    patch_function(transformers.integrations.mxfp4, "load_and_swizzle_mxfp4", load_and_swizzle_mxfp4, match_level = "relaxed")
 
     try:
         from transformers.integrations.mxfp4 import _replace_with_mxfp4_linear
@@ -629,7 +633,7 @@ def patch_GptOssAttention():
         from transformers.models.gpt_oss.modeling_gpt_oss import apply_rotary_pos_emb, repeat_kv
     except Exception as e:
         return raise_error("transformers.models.gpt_oss.modeling_gpt_oss.GptOssAttention", e)
-    
+
     def eager_attention_forward(
         module: nn.Module,
         query: torch.Tensor,
@@ -712,6 +716,7 @@ def patch_GptOssAttention():
         return attn_output, attn_weights
     pass
 
+    functions = []
     def forward(
         self,
         hidden_states: torch.Tensor,
@@ -722,9 +727,7 @@ def patch_GptOssAttention():
         **kwargs: KWARGS_TYPE,
     ) -> tuple[torch.Tensor, Optional[torch.Tensor], Optional[tuple[torch.Tensor]]]:
         return forward_function(self, hidden_states, position_embeddings, attention_mask, past_key_value, cache_position, **kwargs)
-    patch_function(transformers.models.gpt_oss.modeling_gpt_oss.GptOssAttention, "forward", forward)
-
-    # Change past_key_value to past_key_values
+    functions.append(forward)
     def forward(
         self,
         hidden_states: torch.Tensor,
@@ -735,8 +738,8 @@ def patch_GptOssAttention():
         **kwargs: KWARGS_TYPE,
     ) -> tuple[torch.Tensor, Optional[torch.Tensor], Optional[tuple[torch.Tensor]]]:
         return forward_function(self, hidden_states, position_embeddings, attention_mask, past_key_values, cache_position, **kwargs)
-    patch_function(transformers.models.gpt_oss.modeling_gpt_oss.GptOssAttention, "forward", forward)
-
+    functions.append(forward)
+    patch_function_past_key_values(transformers.models.gpt_oss.modeling_gpt_oss.GptOssAttention, "forward", functions)
     # Set env variable for padding purposes
     os.environ["UNSLOTH_ENABLE_FLEX_ATTENTION"] = "1"
 pass
@@ -1061,8 +1064,8 @@ try:
             return raise_error("transformers.models.gpt_oss.configuration_gpt_oss", e)
 
         try:
-            current_class = textwrap.dedent(inspect.getsource(transformers.models.gpt_oss.configuration_gpt_oss.GptOssConfig))
-            new_class = textwrap.dedent(inspect.getsource(Old_GptOssConfig))
+            current_class = dedent(inspect.getsource(transformers.models.gpt_oss.configuration_gpt_oss.GptOssConfig))
+            new_class = dedent(inspect.getsource(Old_GptOssConfig))
             new_class = new_class.replace("Old_GptOssConfig", "GptOssConfig")
             if new_class == current_class:
                 logger.info("Unsloth: Updating GPT OSS Config to fix missing `max_position_embeddings`")
