@@ -131,7 +131,7 @@ class WorkerLoRAManager(AbstractWorkerManager):
                     and model.hf_to_vllm_mapper is not None):
                 hf_to_vllm_mapper = model.hf_to_vllm_mapper
 
-            if lora_request.lora_tensors is not None:
+            if getattr(lora_request, "lora_tensors", None) is not None:
                 lora = self._lora_model_cls.from_lora_tensors(
                     lora_model_id=lora_request.lora_int_id,
                     tensors=lora_request.lora_tensors,
@@ -293,4 +293,4 @@ class LRUCacheWorkerLoRAManager(WorkerLoRAManager):
             loaded = self._adapter_manager.get_adapter(
                 lora_request.lora_int_id) is not None
         self._adapter_manager.activate_adapter(lora_request.lora_int_id)
-        return loaded
\ No newline at end of file
+        return loaded
