@@ -72,6 +72,8 @@ def _return_self(self, *args, **kwargs): return self
 def _return_self_tokenizer(self, *args, **kwargs): return self.tokenizer
 
 def get_target_device(index = 0):
+    if DEVICE_TYPE == "hip":
+        return torch.device("cuda", index)
     return torch.device(DEVICE_TYPE, index)
 
 def get_mem_info():
@@ -1293,6 +1295,8 @@ def load_vllm(
     assert(type(use_bitsandbytes) is bool)
     assert(conservativeness >= 0.0 and conservativeness <= 1.0)
 
+    unsloth_vllm_standby = unsloth_vllm_standby or (os.getenv("UNSLOTH_VLLM_STANDBY", "0") != "0")
+
     if DEVICE_TYPE == "cuda":
         major_version, minor_version = torch.cuda.get_device_capability()
         if major_version < 7: raise NotImplementedError("Unsloth: Your GPU is too old!")
@@ -1330,6 +1334,8 @@ def load_vllm(
 
     # Get correct dtype
     if DEVICE_TYPE == "cuda" and major_version >= 8: _dtype = torch.bfloat16
+    elif DEVICE_TYPE == "hip":
+        _dtype = torch.bfloat16
     elif DEVICE_TYPE == "xpu":
         _dtype = torch.bfloat16
     else:
@@ -1378,6 +1384,8 @@ def load_vllm(
         if (major_version < 7) or (major_version == 7 and minor_version < 5):
             print("Unsloth: Your GPU does not support prefix caching - will disable!")
             enable_prefix_caching = False
+    elif DEVICE_TYPE == "hip":
+        enable_prefix_caching = True
     elif DEVICE_TYPE == "xpu":
         enable_prefix_caching = True
 
